/*
 * PackageLicenseDeclared: Apache-2.0
 * Copyright (c) 2015 ARM Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __MINAR_MINAR_H__
#define __MINAR_MINAR_H__

#include "core-util/Event.h"
#include "core-util/FunctionPointer.h"

namespace minar{

/// @name Types

/// Basic callback type
typedef mbed::util::Event callback_t;

/// Internal time type
typedef uint32_t tick_t;

/// Handle onto scheduled callbacks
typedef void* callback_handle_t;

class SchedulerData;

class Scheduler{
    private:
        class CallbackAdder{
            friend class Scheduler;
            public:
                CallbackAdder& delay(tick_t delay);
                CallbackAdder& tolerance(tick_t tolerance);
                CallbackAdder& period(tick_t tolerance);

                callback_handle_t getHandle();

                ~CallbackAdder();

            private:
                CallbackAdder(Scheduler& sched, callback_t cb);

                Scheduler&   m_sched;
                callback_t   m_cb;
                tick_t       m_tolerance;
                tick_t       m_delay;
                tick_t       m_period;
                bool         m_posted;
        };
    public:
        // get the global scheduler instance
        // The scheduler will be automatically initialised the first time it is
        // referenced.
        // It is not currently possible to de-initialize the scheduler after it
        // has been initialised (it will exist for the lifetime of the
        // program).
        static Scheduler* instance();

        // Function for posting callback with bound argument(s)
        // usage: postCallback([&]{ ...  // }).withDelay(...).requiring(...).releasing(...);
        static CallbackAdder postCallback(callback_t const& cb);

        // Function for posting callbacks using FunctionPointer objects without arguments
        static CallbackAdder postCallback(mbed::util::FunctionPointer& callback)
        {
            return postCallback(callback.bind());
        }

        // Functions for posting callbacks to direct function pointers
        // and objects/member pointers without arguments
        static CallbackAdder postCallback(void (*callback)(void))
        {
            return postCallback(mbed::util::FunctionPointer(callback).bind());
        }

        template<typename T>
        static CallbackAdder postCallback(T *object, void (T::*member)())
        {
            return postCallback(mbed::util::FunctionPointer(object, member).bind());
        }

        static int cancelCallback(callback_handle_t handle);

    private:

        Scheduler();

        // [FPTR] this was a unique_ptr, what's the consequence of making it a simple pointer?
        SchedulerData* data;
};

/// @name Time

/// convert milliseconds into the internal "ticks" time representation
tick_t milliseconds(uint32_t miliseconds);

/// convert ticks to milliseconds
uint32_t ticks(tick_t ticks);


} // namespace minar

#endif // ndef __MINAR_MINAR_H__
