/*
 * PackageLicenseDeclared: Apache-2.0
 * Copyright (c) 2015 ARM Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "cmsis_os.h"
#include "minar/minar.h"

#define EVENT_POOLS_SIZE 50

/// - Private Types

namespace minar{

/// - Memory Pool ID for CallbackNodes
static osPoolId EventPool_Id;

struct CallbackNode {
    CallbackNode(
        minar::callback_t cb,
        minar::tick_t interval,
        os_ptimer ptimer
    ) : cb(cb), interval(interval){
        timer_def.ptimer = ptimer;
        timer_def.timer  = timer_cb;
        timer_id = NULL;
    }

    static void* operator new(std::size_t size){
        (void)size;
        void *p = osPoolAlloc(minar::EventPool_Id);
        if (NULL == p) {
            //CORE_UTIL_RUNTIME_ERROR("Unable to allocate CallbackNode");
        }
        return p;
    }

    static void operator delete(void *p){
        osPoolFree(minar::EventPool_Id, p);
    }

    /// The callback pointer
    minar::callback_t cb;

    /// For more-efficient repeating callbacks, store the interval here and
    /// re-schedule as soon as execution is completed, without another free &
    /// alloc.
    ///
    /// 0 means do not repeat
    minar::tick_t     interval;

    // CMSIS RTOS RTX Timer
    osTimerDef_t      timer_def;
    osTimerId         timer_id;
    uint32_t          timer_cb[6];
};

/// - Memory Pool for CallbackNodes
osPoolDef(EventPool, EVENT_POOLS_SIZE, CallbackNode);

class SchedulerData{
    public:
        SchedulerData();

        minar::callback_handle_t postGeneric(
               // [FPTR] cb below used to be a move ref, is there a better alternative to copy?
               minar::callback_t cb,
               minar::tick_t delay,
               minar::tick_t interval,
               minar::tick_t tolerance
        );

        int cancel(callback_handle_t callback);

        static void timer_callback (void const *argument);
};

/// - Pointer to instance
static minar::Scheduler* staticScheduler = NULL;

} // namespace minar


/// - Implementation of minar class

minar::Scheduler::CallbackAdder& minar::Scheduler::CallbackAdder::delay(
    minar::tick_t delay
){
    m_delay = delay;
    return *this;
}

minar::Scheduler::CallbackAdder& minar::Scheduler::CallbackAdder::tolerance(
    minar::tick_t tolerance
){
    m_tolerance = tolerance;
    return *this;
}

minar::Scheduler::CallbackAdder& minar::Scheduler::CallbackAdder::period(
    minar::tick_t period
){
    m_period = period;
    return *this;
}

minar::callback_handle_t minar::Scheduler::CallbackAdder::getHandle(){
    if(m_cb && !m_posted){
        minar::callback_handle_t temp = m_sched.data->postGeneric(
            // [FPTR] std::move was used below, is there a better way to do this?
            m_cb,
            m_delay,
            m_period,
            m_tolerance
        );
        m_posted = true;
        return temp;
    }
    return NULL;
}

minar::Scheduler::CallbackAdder::~CallbackAdder(){
    getHandle();
}

minar::Scheduler::CallbackAdder::CallbackAdder(Scheduler& sched, callback_t cb)
    : m_sched(sched),
      m_cb(cb),
      m_tolerance(minar::milliseconds(50)),
      m_delay(minar::milliseconds(0)),
      m_period(minar::milliseconds(0)),
      m_posted(false){
}

minar::Scheduler* minar::Scheduler::instance(){
    if(!staticScheduler){
        staticScheduler = new minar::Scheduler();

        minar::EventPool_Id = osPoolCreate(osPool(EventPool));
    }
    return staticScheduler;
}

minar::Scheduler::Scheduler()
    // !!! FIXME: make_unique is C++14
    //: data(std::make_unique<minar::SchedulerData>()){
    : data(new minar::SchedulerData()){
}

minar::Scheduler::CallbackAdder minar::Scheduler::postCallback(
    minar::callback_t const& cb
){
    instance();
    return CallbackAdder(*staticScheduler, cb);
}

int minar::Scheduler::cancelCallback(minar::callback_handle_t handle){
    instance();
    return staticScheduler->data->cancel(handle);
}

/// - SchedulerData Implementation

extern osMessageQId osMessageQId_osTimerMessageQ;

minar::SchedulerData::SchedulerData(){
}

minar::callback_handle_t minar::SchedulerData::postGeneric(
       // [FPTR] cb below used to be a move ref, is there a better alternative to copy?
       minar::callback_t cb,
       minar::tick_t delay,
       minar::tick_t interval,
       minar::tick_t tolerance /* ignored */
){
    CallbackNode* node = new CallbackNode(
        cb,
        interval,
        timer_callback
    );
    if(node) {
        node->timer_id = osTimerCreate(&node->timer_def, osTimerOnce, node);
        if(node->timer_id == NULL) {
            delete node;
            node = NULL;
        }
        if(delay != 0) {
            osTimerStart(node->timer_id, delay);
        } else {
            if(osMessagePut(osMessageQId_osTimerMessageQ, (uint32_t)node->timer_id, 0) != osOK) {
                osTimerDelete(node->timer_id);
                delete node;
                node = NULL;
            }
        }
    }
    return node;
}

int minar::SchedulerData::cancel(minar::callback_handle_t handle) {
    CallbackNode *node = (CallbackNode*)handle;

    if(node) {
        osTimerDelete(node->timer_id);
        delete node;
        return 1;
    }
    return 0;
}

void minar::SchedulerData::timer_callback(void const *argument) {
    CallbackNode* node = (CallbackNode*)argument;

    node->cb();

    if (node->interval != 0) {
        osTimerStart(node->timer_id, node->interval);
    } else {
        osTimerDelete(node->timer_id);
        delete node;
    }
}

/// - Public Function Definitions

/// @name Time

/// convert milliseconds into the internal "ticks" time representation
minar::tick_t minar::milliseconds(uint32_t milliseconds){
    return milliseconds;
}

/// convert ticks into the milliseconds time representation
uint32_t minar::ticks(minar::tick_t ticks){
    return ticks;
}
